#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Constants for mcfonts.

This contains templates for providers, charlists, etc.
Old versions are not stored, and only the latest release is included.
As of this release, 1.19.3 is the latest.

Use :const:`RELEASE_DEFAULT` when looking to template a Vanilla resource without a provider.
"""
from __future__ import annotations

import mcfonts.utils.types

AUTHOR = "WhoAteMyButter"
VERSION = (0, 7)
LICENSE = "MIT"

EMPTY_FONT_JSON: dict[str, list[None]] = {"providers": []}
"""An empty font JSON with no providers."""

RELEASE_DEFAULT: mcfonts.utils.types.TypedProvidersDict = {
    "providers": [
        {"type": "space", "advances": {" ": 4, "\u200c": 0}},
        {
            "type": "bitmap",
            "file": "minecraft:font/nonlatin_european.png",
            "ascent": 7,
            "chars": [
                "¡‰­·₴≠¿×ØÞһðøþΑΒ",
                "ΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣ",
                "ΤΥΦΧΨΩαβγδεζηθικ",
                "λμνξοπρςστυφχψωЂ",
                "ЅІЈЉЊЋАБВГДЕЖЗИК",
                "ЛМНОПРСТУФХЦЧШЩЪ",
                "ЫЬЭЮЯабвгдежзикл",
                "мнопрстуфхцчшщъы",
                "ьэюяєѕіјљњ–—‘’“”",
                "„…⁊←↑→↓⇄＋ƏəɛɪҮүӨ",
                "өʻˌ;ĸẞß₽€ѢѣѴѵӀѲѳ",
                "⁰¹³⁴⁵⁶⁷⁸⁹⁺⁻⁼⁽⁾ⁱ™",
                "ʔʕ⧈⚔☠ҚқҒғҰұӘәҖҗҢ",
                "ңҺאבגדהוזחטיכלמם",
                "נןסעפףצץקר¢¤¥©®µ",
                "¶¼½¾·‐‚†‡•‱′″‴‵‶",
                "‷‹›※‼‽⁂⁈⁉⁋⁎⁏⁑⁒⁗℗",
                "−∓∞☀☁☈Є☲☵☽♀♂⚥♠♣♥",
                "♦♩♪♫♬♭♮♯⚀⚁⚂⚃⚄⚅ʬ⚡",
                "⛏✔❄❌❤⭐⸘⸮⸵⸸⹁⹋⥝ᘔƐ߈",
                "ϛㄥⱯᗺƆᗡƎℲ⅁ꞰꞀԀꝹᴚ⟘∩",
                "Ʌ⅄ɐɔǝɟᵷɥᴉɾʞꞁɯɹʇʌ",
                "ʍʎԱԲԳԴԶԷԹԺԻԼԽԾԿՀ",
                "ՁՂՃՄՅՆՇՈՉՋՌՍՎՏՐՑ",
                "ՒՓՔՕՖՙաբգդեզէըթժ",
                "իլխծկհձղճմյնշոչպ",
                "ջռսվտրցւփքօֆևשתԸ",
                "՚՛՜՝՞՟ՠֈ֏¯ſƷʒǷƿȜ",
                "ȝȤȥ˙Ꝛꝛ‑⅋⏏⏩⏪⏭⏮⏯⏴⏵",
                "⏶⏷⏸⏹⏺⏻⏼⏽⭘▲▶▼◀●◦◘",
                "⚓⛨ĲĳǉꜨꜩꜹꜻﬀﬁﬂﬃﬅ�Ե",
                "Պᚠᚢᚣᚤᚥᚦᚧᚨᚩᚪᚫᚬᚭᚮᚯ",
                "ᚰᚱᚲᚳᚴᚶᚷᚸᚹᚺᚻᚼᚽᚾᚿᛀ",
                "ᛁᛂᛃᛄᛅᛆᛇᛈᛉᛊᛋᛌᛍᛎᛏᛐ",
                "ᛑᛒᛓᛔᛕᛖᛗᛘᛙᛚᛛᛜᛝᛞᛟᛠ",
                "ᛡᛢᛣᛤᛥᛦᛧᛨᛩᛪ᛫᛬᛭ᛮᛯᛰ",
                "ᛱᛲᛳᛴᛵᛶᛷᛸ☺☻¦☹ך׳״װ",
                "ױײ־׃׆´¨ᴀʙᴄᴅᴇꜰɢʜᴊ",
                "ᴋʟᴍɴᴏᴘꞯʀꜱᴛᴜᴠᴡʏᴢ§",
                "ɱɳɲʈɖɡʡɕʑɸʝʢɻʁɦʋ",
                "ɰɬɮʘǀǃǂǁɓɗᶑʄɠʛɧɫ",
                "ɨʉʊɘɵɤɜɞɑɒɚɝƁƉƑƩ",
                "ƲႠႡႢႣႤႥႦႧႨႩႪႫႬႭႮ",
                "ႯႰႱႲႳႴႵႶႷႸႹႺႻႼႽႾ",
                "ႿჀჁჂჃჄჅჇჍაბგდევზ",
                "თიკლმნოპჟრსტუფქღ",
                "ყშჩცძწჭხჯჰჱჲჳჴჵჶ",
                "ჷჸჹჺ჻ჼჽჾჿתּשׂפֿפּכּײַיִ",
                "וֹוּבֿבּꜧꜦɺⱱʠʗʖɭɷɿʅʆ",
                "ʓʚ₪₾֊ⴀⴁⴂⴃⴄⴅⴆⴡⴇⴈⴉ",
                "ⴊⴋⴌⴢⴍⴎⴏⴐⴑⴒⴣⴓⴔⴕⴖⴗ",
                "ⴘⴙⴚⴛⴜⴝⴞⴤⴟⴠⴥ⅛⅜⅝⅞⅓",
                "⅔✉☂☔☄⛄☃⌛⌚⚐✎❣♤♧♡♢",
                "⛈☰☱☳☴☶☷↔⇒⇏⇔⇵∀∃∄∉",
                "∋∌⊂⊃⊄⊅∧∨⊻⊼⊽∥≢⋆∑⊤",
                "⊥⊢⊨≔∁∴∵∛∜∂⋃⊆⊇□△▷",
                "▽◁◆◇○◎☆★✘₀₁₂₃₄₅₆",
                "₇₈₉₊₋₌₍₎∫∮∝⌀⌂⌘〒ɼ",
                "ƄƅẟȽƚƛȠƞƟƧƨƪƸƹƻƼ",
                "ƽƾȡȴȵȶȺⱥȻȼɆɇȾⱦɁɂ",
                "ɃɄɈɉɊɋɌɍɎɏẜẝỼỽỾỿ",
                "Ꞩꞩ𐌰𐌱𐌲𐌳𐌴𐌵𐌶𐌷𐌸𐌹𐌺𐌻𐌼𐌽",
                "𐌾𐌿𐍀𐍁𐍂𐍃𐍄𐍅𐍆𐍇𐍈𐍉𐍊🌧🔥🌊",
                "⅐⅑⅕⅖⅗⅙⅚⅟↉🗡🏹🪓🔱🎣🧪⚗",
                "⯪⯫Ɑ🛡✂🍖🪣🔔⏳⚑₠₡₢₣₤₥",
                "₦₩₫₭₮₰₱₲₳₵₶₷₸₹₺₻",
                "₼₿              ",
            ],
        },
        {
            "type": "bitmap",
            "file": "minecraft:font/accented.png",
            "height": 12,
            "ascent": 10,
            "chars": [
                "ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏ",
                "ÐÑÒÓÔÕÖÙÚÛÜÝàáâã",
                "äåæçìíîïñòóôõöùú",
                "ûüýÿĀāĂăĄąĆćĈĉĊċ",
                "ČčĎďĐđĒēĔĕĖėĘęĚě",
                "ĜĝḠḡĞğĠġĢģĤĥĦħĨĩ",
                "ĪīĬĭĮįİıĴĵĶķĹĺĻļ",
                "ĽľĿŀŁłŃńŅņŇňŊŋŌō",
                "ŎŏŐőŒœŔŕŖŗŘřŚśŜŝ",
                "ŞşŠšŢţŤťŦŧŨũŪūŬŭ",
                "ŮůŰűŲųŴŵŶŷŸŹźŻżŽ",
                "žǼǽǾǿȘșȚțΆΈΉΊΌΎΏ",
                "ΐΪΫάέήίΰϊϋόύώЀЁЃ",
                "ЇЌЍЎЙйѐёђѓїћќѝўџ",
                "ҐґḂḃḊḋḞḟḢḣḰḱṀṁṖṗ",
                "ṠṡṪṫẀẁẂẃẄẅỲỳèéêë",
                "ŉǧǫЏḍḥṛṭẒỊịỌọỤụ№",
                "ȇƔɣʃ⁇ǱǲǳǄǅǆǇǈǊǋǌ",
                "ℹᵫꜲꜳꜴꜵꜶꜷꜸꜺꜼꜽꝎꝏꝠꝡ",
                "ﬄﬆᚡᚵƠơƯưẮắẤấẾếốỚ",
                "ớỨứẰằẦầỀềồỜờỪừẢả",
                "ẲẳẨẩẺẻổỞỂểỈỉỎỏỔở",
                "ỦủỬửỶỷẠạẶặẬậẸẹỆệ",
                "ỘộỢợỰựỴỵỐƕẪẫỖỗữ☞",
                "☜☮ẴẵẼẽỄễỒỠỡỮỸỹҘҙ",
                "ҠҡҪҫǶ⚠⓪①②③④⑤⑥⑦⑧⑨",
                "⑩⑪⑫⑬⑭⑮⑯⑰⑱⑲⑳ⒶⒷⒸⒹⒺ",
                "ⒻⒼⒽⒾⒿⓀⓁⓂⓃⓄⓅⓆⓇⓈⓉⓊ",
                "ⓋⓌⓍⓎⓏⓐⓑⓒⓓⓔⓕⓖⓗⓘⓙⓚ",
                "ⓛⓜⓝⓞⓟⓠⓡⓢⓣⓤⓥⓦⓧⓨⓩ̧",
                "ʂʐɶǍǎǞǟǺǻȂȃȦȧǠǡḀ",
                "ḁȀȁḆḇḄḅᵬḈḉḐḑḒḓḎḏ",
                "ḌᵭḔḕḖḗḘḙḜḝȨȩḚḛȄȅ",
                "ȆᵮǴǵǦḦḧḨḩḪḫȞȟḤẖḮ",
                "ḯȊȋǏǐȈȉḬḭǰȷǨǩḲḳḴ",
                "ḵḺḻḼḽḶḷḸḹⱢḾḿṂṃᵯṄ",
                "ṅṆṇṊṋǸǹṈṉᵰǬǭȬȭṌṍ",
                "ṎṏṐṑṒṓȎȏȪȫǑǒȮȯȰȱ",
                "ȌȍǪṔṕᵱȒȓṘṙṜṝṞṟȐȑ",
                "ṚᵳᵲṤṥṦṧṢṣṨṩᵴṰṱṮṯ",
                "ṬẗᵵṲṳṶṷṸṹṺṻǓǔǕǖǗ",
                "ǘǙǚǛǜṴṵȔȕȖṾṿṼṽẆẇ",
                "ẈẉẘẌẍẊẋȲȳẎẏẙẔẕẐẑ",
                "ẓᵶǮǯẛꜾꜿǢǣᵺỻᴂᴔꭣȸʣ",
                "ʥʤʩʪʫȹʨʦʧꭐꭑ₧Ỻאַאָƀ",
                "ƂƃƇƈƊƋƌƓǤǥƗƖɩƘƙƝ",
                "ƤƥɽƦƬƭƫƮȗƱƜƳƴƵƶƢ",
                "ƣȢȣʭʮʯﬔﬕﬗﬖﬓӐӑӒӓӶ",
                "ӷҔҕӖӗҼҽҾҿӚӛӜӝӁӂӞ",
                "ӟӢӣӤӥӦӧӪӫӰӱӮӯӲӳӴ",
                "ӵӸӹӬӭѶѷӔӺԂꚂꚀꚈԪԬꚄ",
                "ԄԐӠԆҊӃҞҜԞԚӅԮԒԠԈԔ",
                "ӍӉԨӇҤԢԊҨԤҦҎԖԌꚐҬꚊ",
                "ꚌԎҲӼӾԦꚔҴꚎҶӋҸꚒꚖꚆҌ",
                "ԘԜӕӻԃꚃꚁꚉԫԭꚅԅԑӡԇҋ",
                "ӄҟҝԟԛӆԯԓԡԉԕӎӊԩӈҥ",
                "ԣԋҩԥҧҏԗԍꚑҭꚋꚍԏҳӽӿ",
                "ԧꚕҵꚏҷӌҹꚓꚗꚇҍԙԝἈἀἉ",
                "ἁἊἂἋἃἌἄἍἅἎἆἏἇᾺὰᾸ",
                "ᾰᾹᾱΆάᾈᾀᾉᾁᾊᾂᾋᾃᾌᾄᾍ",
                "ᾅᾎᾆᾏᾇᾼᾴᾶᾷᾲᾳἘἐἙἑἚ",
                "ἒἛἓἜἔἝἕῈΈὲέἨἠῊὴἩ",
                "ἡἪἢἫἣἬἤἭἥἮἦἯἧᾘᾐᾙ",
                "ᾑᾚᾒᾛᾓᾜᾔᾝᾕᾞᾖᾟᾗΉήῌ",
                "ῃῂῄῆῇῚὶΊίἸἰἹἱἺἲἻ",
                "ἳἼἴἽἵἾἶἿἷῘῐῙῑῒΐῖ",
                "ῗῸὸΌόὈὀὉὁὊὂὋὃὌὄὍ",
                "ὅῬῤῥῪὺΎύὙὑὛὓὝὕὟὗ",
                "ῨῠῩῡϓϔῢΰῧὐὒὔῦὖῺὼ",
                "ΏώὨὠὩὡὪὢὫὣὬὤὭὥὮὦ",
                "Ὧὧᾨᾠᾩᾡᾪᾢᾫᾣᾬᾤᾭᾥᾮᾦ",
                "ᾯᾧῼῳῲῴῶῷ☯☐☑☒ƍƺⱾȿ",
                "ⱿɀᶀꟄꞔᶁᶂᶃꞕᶄᶅᶆᶇᶈᶉᶊ",
                "ᶋᶌᶍꟆᶎᶏᶐᶒᶓᶔᶕᶖᶗᶘᶙᶚ",
                "ẚ⅒⅘₨₯           ",
            ],
        },
        {
            "type": "bitmap",
            "file": "minecraft:font/ascii.png",
            "ascent": 7,
            "chars": [
                "                ",
                "                ",
                " !\"#$%&'()*+,-./",
                "0123456789:;<=>?",
                "@ABCDEFGHIJKLMNO",
                "PQRSTUVWXYZ[\\]^_",
                "`abcdefghijklmno",
                "pqrstuvwxyz{|}~ ",
                "                ",
                "            £  ƒ",
                "      ªº  ¬   «»",
                "░▒▓│┤╡╢╖╕╣║╗╝╜╛┐",
                "└┴┬├─┼╞╟╚╔╩╦╠═╬╧",
                "╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀",
                "             ∅∈ ",
                "≡±≥≤⌠⌡÷≈°∙ √ⁿ²■ ",
            ],
        },
        {
            "type": "legacy_unicode",
            "sizes": "minecraft:font/glyph_sizes.bin",
            "template": "minecraft:font/unicode_page_%s.png",
        },
    ]
}
"""The font JSON of default.json, the default Minecraft font."""
RELEASE_ALT: mcfonts.utils.types.TypedProvidersDict = {
    "providers": [
        {
            "type": "bitmap",
            "file": "minecraft:font/ascii_sga.png",
            "ascent": 7,
            "chars": [
                "                ",
                "                ",
                "                ",
                "                ",
                " ABCDEFGHIJKLMNO",
                "PQRSTUVWXYZ     ",
                " abcdefghijklmno",
                "pqrstuvwxyz     ",
                "                ",
                "                ",
                "                ",
                "                ",
                "                ",
                "                ",
                "                ",
                "                ",
            ],
        }
    ]
}
"""The font JSON of alt.json, the SGA font."""
RELEASE_UNIFORM: mcfonts.utils.types.TypedProvidersDict = {
    "providers": [
        {
            "type": "legacy_unicode",
            "sizes": "minecraft:font/glyph_sizes.bin",
            "template": "minecraft:font/unicode_page_%s.png",
        }
    ]
}
"""The font JSON of uniform.json, the Unicode fallback font."""

PROVIDER_NONLATIN = RELEASE_DEFAULT["providers"][1]
"""The default provider for nonlatin-european."""
PROVIDER_ACCENTED = RELEASE_DEFAULT["providers"][2]
"""The default provider for accented."""
PROVIDER_ASCII = RELEASE_DEFAULT["providers"][3]
"""The default provider for ASCII."""

PADDING_CHARS = {"\0", " "}
"""Characters that act as padding; glyphs cannot be assigned to these chars."""

UNKNOWN_FIELD = "???"
"""A fallback magic string."""



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
**mcfonts** is a versatile, fast, and extensible package for working with Minecraft fonts.

mcfonts works with any valid font JSON and can export every kind of texture and sizing.

| For more information, see `<https://gitlab.com/whoatemybutter/mcfonts/>`_.
| Read the documentation online at `<https://mcfonts.rtfd.io>`_.

----

| Licensed under the MIT license, see https://choosealicense.com/licenses/mit/ for details.
| Formatted with Black, see https://github.com/psf/black.
| Checked with Pylint, see https://pylint.org/.
"""
from __future__ import annotations

import collections.abc
import json
import logging
import math
import os
import os.path
import pathlib
import shutil
import sys
import tempfile
import typing
import warnings
import zipfile

import PIL.Image
import tinyunicodeblock
import unirange

import mcfonts.compacting
import mcfonts.constants
import mcfonts.coverage_reports
import mcfonts.exceptions
import mcfonts.export_formats
import mcfonts.filters
import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.providers.bitmap
import mcfonts.providers.legacy_unicode
import mcfonts.providers.reference
import mcfonts.providers.space
import mcfonts.providers.ttf
import mcfonts.providers.unihex
import mcfonts.utils
import mcfonts.utils.importing
import mcfonts.utils.resources

__author__ = mcfonts.constants.AUTHOR
__version__ = mcfonts.constants.VERSION
__license__ = mcfonts.constants.LICENSE

if sys.version_info < (3, 11, 0):
    raise RuntimeError(f"minimum Python version is 3.11.0, you are running {sys.version.split(' ', 1)[0]}")

# Decompression bombs will error, as they should
warnings.simplefilter("error", PIL.Image.DecompressionBombWarning)

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    format="[%(relativeCreated)d] [%(name)s/%(levelname)s]: (at %(funcName)s()) %(message)s",
)


class MinecraftFont:
    """
    The MinecraftFont class.

    Requires the providers of a provider file, and the associated resources mapping.

    You should never instantiate this class directly.
    Use :mod:`mcfonts.importing`.

    If you need to add, modify, or remove providers, do it through :attr:`self.providers`;
    it's a list of Provider classes, each containing relevant fields and methods.

    Be sure to run :meth:`mcfonts.MinecraftFont.validate` after making any changes;
    it won't be done automatically.

    .. warning::
        If more than one :class:`mcfonts.providers.OptionsProvider` is present in `provider_list`,
        only the **last** one will be used.

    :param provider_list:
        A list of providers, all of which are instances of :data:`mcfonts.providers.base.Provider`.
    :param is_monochrome:
        Whether font resources are loaded in grayscale or not.
        Default is True.
    """

    def __init__(
        self,
        provider_list: list[mcfonts.providers.base.Provider],
        is_monochrome: bool = True,
    ):
        self.providers = provider_list
        """Do not iterate over this! Use :meth:`yield_providers` instead."""
        self.is_monochrome = is_monochrome

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.providers}, {self.is_monochrome})"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __hash__(self) -> int:
        return hash((self.providers, self.is_monochrome))

    def __add__(self, other: MinecraftFont) -> MinecraftFont:
        return MinecraftFont(self.providers, self.is_monochrome) + other

    def __iadd__(self, other: MinecraftFont) -> MinecraftFont:
        self.providers += other.providers
        if other.is_monochrome is False:
            self.is_monochrome = False
        self.validate()
        return self

    def __len__(self) -> int:
        return len(list(self.yield_characters()))

    def export(
        self,
        export_format: collections.abc.Callable[
            [dict[str, mcfonts.glyphs.Glyph], mcfonts.export_formats.ExportSettings], bytes
        ],
        settings: mcfonts.export_formats.ExportSettings | None = None,
        filter_instance: mcfonts.filters.Filter | None = None,
    ) -> bytes:
        """
        Export this Minecraft font to another format.
        See available formats at :const:`mcfonts.export_formats.EXPORT_FORMATS_MAP`.

        :param export_format:
            The function to pass glyphs and settings to.
            Look in :mod:`mcfonts.export_formats` for a list of the functions you can pass.
            Also accepts a string of the format name (see :const:`mcfonts.export_formats.EXPORT_FORMATS_MAP`).
        :param settings:
            A dictionary of the needed parameters to attach to the exported font.
            See :class:`mcfonts.export_formats.ExportSettings`.
        :param filter_instance:
            An optional instance of :class:`mcfonts.filters.Filter`.
            If supplied, filter rules will be obeyed.
        :returns: The exported data in the specified format that can be written to a file and saved.
        """
        cached_glyphs: dict[str, mcfonts.glyphs.Glyph] = {}
        num_providers = 0
        for provider in self.yield_providers():
            if filter_instance and filter_instance.check_provider(provider.__class__):
                logger.info(f"Skipping provider {num_providers + 1:,} {provider.pretty_print()} (filtered)")
                continue
            for character, glyph in provider.yield_glyphs():
                if mcfonts.utils.unicode.is_codepoint_surrogate(ord(character)) or (
                    filter_instance and filter_instance.check_character(character)
                ):  # Don't include surrogates or filtered characters.
                    continue
                if glyph:
                    cached_glyphs[character] = glyph
            num_providers += 1
        return export_format(
            cached_glyphs,
            settings
            or {
                "name": None,
                "author": None,
                "license": None,
                "include_credits": None,
            },
        )

    def yield_providers(
        self, unwind_reference_children: bool = True
    ) -> collections.abc.Iterable[mcfonts.providers.base.Provider]:
        """
        Yield the providers in this font.

        By default, this will unpack reference providers and yield its children instead.
        Reference providers will never yield the same referenced font twice if `unwind_reference_children` is True.

        :param unwind_reference_children:
            Whether to yield a reference provider's children instead of the reference provider plainly.
            Usually, this is a good idea.
            This also prevents the same font from being yielded twice.
        :returns: A yield of Provider.
        """
        referenced_providers: list[str] = []
        for provider in self.providers:
            if isinstance(provider, mcfonts.providers.reference.ReferenceProvider) and unwind_reference_children:
                assert isinstance(provider, mcfonts.providers.reference.ReferenceProvider)
                # Reference providers require no duplication, so store each font ID and check it.
                if (referenced := provider.contents["id"]) in referenced_providers:
                    # It's already been yielded, ignore it now.
                    continue
                # Not yielded yet.
                referenced_providers.append(referenced)
                yield from provider.yield_children()
            else:
                yield provider

    def yield_characters(self) -> collections.abc.Iterable[str]:
        """
        Yield all the characters this font covers and has a definition for.

        .. note:: Any character in :const:`mcfonts.constants.PADDING_CHARS` isn't counted.

        :returns: A yield of strings of individual characters.
        """
        for provider in self.yield_providers():
            yield from provider.yield_characters()

    def save(self, indent: int | str | None = "\t") -> None:
        """
        Recursively write all providers to their original file locations.

        This is different from exporting.
        The file is indented by default.

        .. warning:: Not to be confused with :meth:`MinecraftFont.export`.

        .. warning:: Doesn't save resources.

        :param indent:
            The indentation level, refer to :func:`json.dump()` for possible values.
        """
        origin_cache: dict[str, list[mcfonts.utils.types.TypedProviderDict]] = {}
        providers_total: list[mcfonts.providers.base.Provider] = []
        for wound_provider in self.yield_providers(False):
            providers_total.append(wound_provider)
            if isinstance(wound_provider, mcfonts.providers.reference.ReferenceProvider):
                providers_total.extend(wound_provider.yield_children())

        for provider in providers_total:
            if (origin_str := str(provider.origin)) not in origin_cache:
                origin_cache[origin_str] = []
            origin_cache[origin_str].append(provider.get_contents())

        for path, contents in origin_cache.items():
            with open(path, "w+", encoding="utf-8") as open_path:
                json.dump({"providers": contents}, open_path, ensure_ascii=False, indent=indent)

    def count_providers(self) -> dict[str, int]:
        """
        Return a counted summary of the providers this font contains.

        This is future-proof, and will work with any provider as long as it has a "type" key.

        :returns: A summary of font's providers.
        """
        result = {}
        for provider in self.yield_providers():
            if (provider_type := provider.provider_type) not in result:
                result[provider_type] = 1
            else:
                result[provider_type] += 1
        return result

    def count_providers_total(self) -> int:
        """
        Count the number of providers in the font.

        :returns: Number of providers.
        """
        return len(self.count_providers())

    def print_info(self, table_chars: bool = True, summary_only: bool = False) -> None:
        """
        Print basic information about the font.

        :param table_chars:
            Whether to print a 'chars' list as a square table, or as a simple string.
            This only applies to :class:`mcfonts.providers.BitmapProvider`.
        :param summary_only:
            If True, will only print the number of characters and providers.
        """
        if not summary_only:
            for provider in self.yield_providers():
                if isinstance(provider, mcfonts.providers.bitmap.BitmapProvider):
                    assert isinstance(provider, mcfonts.providers.bitmap.BitmapProvider)
                    provider.print_info(table_chars)
                else:
                    provider.print_info()
            print("\n")
        print(f"Characters: {len(list(self.yield_characters())):,}")
        print(f"Providers: {self.count_providers_total():,}")

    def validate(self) -> None:
        """Run basic structure checks on the providers of the font JSON."""
        if len(self.providers) < 1:
            logger.warning("There are no providers.")
        for provider in self.yield_providers():
            if isinstance(provider, mcfonts.providers.base.Provider):
                provider.validate()
            else:
                raise mcfonts.exceptions.ProviderError(
                    mcfonts.providers.base.format_provider_message(provider, "is not a valid provider.")
                )

    def compact(
        self,
        chars_in_row: int = 0,
        cell_size: tuple[int, int] = (0, 0),
        square_cells: bool = True,
        output_file: str | None = None,
    ) -> tuple[list[str], PIL.Image.Image, tuple[int, int]]:
        """
        Take all "bitmap" providers and export every character sheet into a single sheet.

        Characters are scaled according to the largest effective bounding box in all providers.

        :param chars_in_row:
            How many characters to fit inside each row of the resulting sheet.
            If this is 0, this will be set to the length of the first string in the
            "charlist" list. If this is negative, this will be set so that the resulting sheet is
            square. By default, this is 0 (auto first string).
        :param cell_size:
            What size to make each glyph cell.
            If this is (0, 0), this will be set to the largest dimensions of every glyph in `glyphs`.
        :param square_cells:
            If True, each glyph's width will equal its height.
            This is based on whichever number is largest.
            If False, each glyph's width will be unrelated to its height.
        :param output_file: Where to write the sheet to. If this is None, nothing will be
            written.
        :returns: A list of the new characters, and the new file as a :class:`PIL.Image.Image`.
        """
        extracted: dict[str, mcfonts.glyphs.BitmapGlyph | mcfonts.glyphs.UnihexGlyph] = {}
        for provider in self.yield_providers():
            if isinstance(provider, (mcfonts.providers.bitmap.BitmapProvider, mcfonts.providers.unihex.UnihexProvider)):
                yields = provider.yield_glyphs()
                for character, glyph in yields:
                    if glyph:
                        extracted[character] = glyph

        if chars_in_row == 0:
            bitmap_providers = [
                provider
                for provider in self.yield_providers()
                if isinstance(provider, mcfonts.providers.bitmap.BitmapProvider)
            ]

            chars_in_row = max(len(provider.contents["chars"][0] or "") for provider in bitmap_providers)
        elif chars_in_row < 0:
            chars_in_row = math.ceil(math.sqrt(len(extracted)))

        compacted = mcfonts.compacting.compact_glyphs(list(extracted.values()), chars_in_row, cell_size, square_cells)
        sheet = (
            list(mcfonts.providers.bitmap.fit_chars_into_charlist(list(extracted.keys()), chars_in_row)),
            compacted[0],
            compacted[1],
        )
        if output_file:
            with open(mcfonts.utils.resources.expand_path(output_file), "wb") as open_output_file:
                sheet[1].save(open_output_file)
        return sheet

    def coverage_report(self) -> mcfonts.coverage_reports.CoverageReport:
        """
        Build a report of what characters this font contains.

        This includes information like how many characters are in the font,
        and what Unicode blocks are covered.

        :returns: A dictionary of ``{"chars": int, "blocks": {str: int}}``.
        """
        characters: list[str] = []
        blocks: dict[str, int] = {}
        for character in self.yield_characters():
            characters.append(character)
            if (block := tinyunicodeblock.block(character, include_csur=True)) in blocks:
                blocks[block] += 1
            else:
                blocks[block] = 1
        return mcfonts.coverage_reports.CoverageReport(set(characters), blocks)

    def yield_glyphs_in_unirange(
        self, unirange_notation: str
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph | None]]:
        """
        Given a `unirange_notation`, yield a tuple of the requested character to their glyph.

        :param unirange_notation:
            A string representing the requested range of chars.
            See https://pypi.org/project/unirange/.
        :returns: A yield of the requested characters and associated glyphs that match `unirange_notation`.
        """
        for provider in self.yield_providers():
            yield from provider.yield_glyphs_in_unirange(unirange_notation)

    def get_covering_providers(self, unirange_notation: str) -> list[mcfonts.providers.base.Provider]:
        """
        Given a codepoint range, return a list of providers that cover these characters.

        :param unirange_notation:
            A string representing the requested range of chars.
            See https://pypi.org/project/unirange/.
        :returns: A list of the providers that cover codeopints defined in `unirange_notation`.
        """
        result = []
        covers = unirange.unirange_to_characters(unirange_notation)
        for provider in self.yield_providers():
            if not isinstance(provider, mcfonts.providers.space.SpaceProvider):
                # Ignore padding chars
                covers.difference_update(mcfonts.constants.PADDING_CHARS)
            if provider.chars_covered.intersection(covers):
                result.append(provider)
        return result

    def reload_to_monochrome(self) -> None:
        """
        Replace the resources used in the providers with a grayscale version.

        If the resource is already grayscale, this will have no effect.
        This modifies the resource of this provider in place, and **can't be undone**.
        """
        if self.is_monochrome:
            mcfonts.logger.info("Font is already in monochrome; can't reload")
            return
        for provider in self.yield_providers():
            if isinstance(provider, mcfonts.providers.bitmap.BitmapProvider):
                provider.reload_to_monochrome()
        self.is_monochrome = True

    def compare(self, other: typing.Self) -> None:
        """
        Given `other`, compare the two, using `self` as a baseline.

        The information compared is:

        * Character count
        * Blocks covered
        * Providers included

        :param other: A second instance of :class:`mcfonts.MinecraftFont` to compare to.
        """
        self.coverage_report().compare(other.coverage_report())
        if mcfonts.utils.colors.USE_COLORS:
            print(f"\n{mcfonts.utils.colors.BRIGHT}PROVIDERS{mcfonts.utils.colors.RESET_ALL}")
        else:
            print("\nPROVIDERS")
        print(":: type: this | other (delta)")
        providers_this = {"bitmap": 0, "space": 0, "ttf": 0, "legacy_unicode": 0}
        providers_other = providers_this.copy()
        for provider in self.yield_providers():
            providers_this[provider.provider_type] += 1
        for provider in other.yield_providers():
            providers_other[provider.provider_type] += 1
        for provider_type in ("bitmap", "space", "ttf", "legacy_unicode"):
            amount_this = providers_this[provider_type]
            amount_other = providers_other[provider_type]
            print(
                f"\t{provider_type}: "
                f"{amount_this} | {amount_other} "
                f"({mcfonts.utils.colors.color_number(amount_other - amount_this)})"
            )

    def regenerate_charlists(self) -> mcfonts.utils.types.TypedProvidersDict:
        """
        Iterate through each resource and analyse it.

        For each resource, an entry will be made in a dictionary that contains the same height and ascent fields as
        the original, except the "chars" list will be updated to reflect the glyphs in the associated resource.

        .. warning::
            This only works if the first character in the original charlist is present,
            **and** if the order is incremental (U+0100, then U+0101, then U+0102, so on).

            If this condition isn't true, this function **will** fail.
            There are **no** checks for this rule.

        :returns: A dictionary matching a normal font JSON, with each "chars" list updated.
        """
        result: list[mcfonts.utils.types.TypedProviderDict] = []
        for provider in self.yield_providers():
            if isinstance(provider, mcfonts.providers.bitmap.BitmapProvider) and provider.resource:
                old_contents = provider.get_contents().copy()
                old_contents["chars"] = mcfonts.providers.bitmap.resource_to_charlist(
                    provider.resource,
                    chr(int(os.path.basename(provider.contents["file"]).split("-", 1)[0].split(".", 1)[0], 16)),
                    provider.glyph_cell,
                )
                result.append(old_contents)
            else:
                result.append(provider.get_contents())
        return {"providers": result}


def from_java_font_file(
    file_path: pathlib.Path,
    read_colors: bool = False,
    strict: bool = True,
) -> mcfonts.MinecraftFont:
    """
    Load a Java Edition font JSON into a :class:`mcfonts.MinecraftFont`.

    Requires a "providers" list, and missing files will raise an error.

    :param file_path:
        The real path to the font JSON.
    :param read_colors:
        If True, glyph will be loaded in RGBA.
        If false, loaded in LA.

        RGBA images incur **heavy** time cost.
        Be careful.
    :param strict:
        If True:

        * Bad providers will raise an error.
        * Missing files will raise an error.

        If False:

        * Bad providers will be ignored.
        * Missing files will be skipped.

    :returns: A :class:`mcfonts.MinecraftFont` instance.
    :raises FileNotFoundError: If a referenced file is not found and `strict` is True.
    """
    with open(file_path, encoding="utf-8") as datafp:
        file_contents: dict[str, typing.Any] = json.load(datafp, strict=False)
    return from_java_font_contents(file_contents, file_path, read_colors, strict)


def from_java_font_contents(
    file_contents: dict[str, typing.Any],
    file_path: pathlib.Path,
    read_colors: bool = False,
    strict: bool = True,
) -> mcfonts.MinecraftFont:
    """
    Load a Java Edition font JSON into a :class:`mcfonts.MinecraftFont`.

    Requires a "providers" list, and missing files will raise an error.

    :param file_contents:
        The contents of the font JSON file, loaded as a dictionary.
        This dictionary should include the base "providers" key.
    :param file_path:
        The path to the font JSON.
        This is needed for loading resources.
    :param read_colors:
        If True, glyph will be loaded in RGBA (Red, Green, Blue, Alpha). If false, loaded in LA.
        RGBA images incur **heavy** time cost. Be careful.
    :param strict:
        If True:

        * Bad providers will raise an error.
        * Missing files will raise an error.
        If False:

        * Bad providers will be ignored.
        * Missing files will be skipped.

    :returns: A :class:`mcfonts.MinecraftFont` instance.
    :raises FileNotFoundError: If a referenced file isn't found and `strict` is True.
    """
    provider_content = file_contents.get("providers", [])
    if read_colors:
        mode = "RGBA"
    else:
        mode = "LA"
    return mcfonts.MinecraftFont(
        list(
            mcfonts.utils.importing.load_all(
                provider_content, mcfonts.utils.resources.expand_path(file_path), strict, mode
            )
        ),
        not read_colors,
    )


def from_java_pack_folder(
    folder_path: pathlib.Path,
    font_file_name: str = "default.json",
    namespace: str = "minecraft",
    read_colors: bool = False,
    strict: bool = True,
) -> mcfonts.MinecraftFont:
    """
    Load a Java Edition resource pack into a :class:`mcfonts.MinecraftFont`.

    The font must be in the path ``assets/<namespace>/font/<fontfile>``.

    :param folder_path:
        The real path to the folder that contains a resource pack.
        This isn't the ``/assets/`` folder, nor is it a ZIP file.
        The files inside this folder should be ``assets/``, and ``pack.mcmeta``.
    :param font_file_name:
        The name of the font file.
        By default, this is "default.json".
    :param namespace:
        The namespace to find assets in.
        By default, this is "minecraft".
    :param read_colors:
        If True, glyph will be loaded in 'RGBA'.
        If False, loaded in 'LA'.

        RGBA images incur **heavy** time cost.
        Be careful.
    :param strict:
        If True:

        * Bad providers will raise an error.
        * Missing files will raise an error.

        If False:

        * Bad providers will be ignored.
        * Missing files will be skipped.

    :returns: A :class:`mcfonts.MinecraftFont` instance.
    :raises FileNotFoundError: If a referenced file isn't found and `strict` is True.
    """

    return from_java_font_file(
        mcfonts.utils.resources.expand_path(folder_path).joinpath(
            f"assets/{namespace}/font/{font_file_name}",
        ),
        read_colors,
        strict,
    )


def from_java_pack_zip(
    file_path: pathlib.Path, password: bytes | None = None, read_colors: bool = False, strict: bool = True
) -> mcfonts.MinecraftFont:
    """
    Load a Java Edition resource pack ZIP into a :class:`mcfonts.MinecraftFont`.

    The font must be in the path ``assets/<namespace>/font/<fontfile>``.

    :param file_path:
        The real path to the ZIP file.
    :param password:
        Password to use when reading the ZIP file.
        Set to ``None`` if there is no password.
    :param read_colors:
        If True, glyph will be loaded in 'RGBA'.
        If False, loaded in 'LA'.

        RGBA images incur **heavy** time cost.
        Be careful.
    :param strict:
        If True:

        * Bad providers will raise an error.
        * Missing files will raise an error.

        If False:

        * Bad providers will be ignored.
        * Missing files will be skipped.

    :returns: A :class:`mcfonts.MinecraftFont` instance.
    :raises FileNotFoundError: If a referenced file isn't found and `strict` is True.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        real_temp_dir = pathlib.Path(temp_dir).absolute()
        with zipfile.ZipFile(mcfonts.utils.resources.expand_path(file_path)) as zip_file:
            zip_file.extractall(real_temp_dir, pwd=password)
        return from_java_pack_folder(real_temp_dir, read_colors=read_colors, strict=strict)


def from_java_resource_template(
    file_path: pathlib.Path,
    template_provider: dict[str, typing.Any],
    read_colors: bool = False,
    strict: bool = True,
) -> mcfonts.MinecraftFont:
    """
    | Given the path to a texture and the contents of an individual font provider,
    | return a :class:`mcfonts.MinecraftFont` instance with it, and the resource in `file_path`.

    ``template_provider["file"]`` can be any value; it will be overwritten anyway,
    although it must exist.

    :param file_path:
        The path to the PNG :term:`resource` that needs templating.
    :param template_provider:
        An individual provider dictionary.
        Not a list of providers.
    :param read_colors:
        If True, glyph will be loaded in 'RGBA'.
        If False, loaded in 'LA'.
        RGBA images incur **heavy** time cost. Be careful.
    :param strict:
        If a provider has bad data,
        an exception will be raised and no provider list will be returned if this is True.
        If this is False, it will be ignored.
    :returns:
        A :class:`mcfonts.MinecraftFont` instance.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        # Put the PNG file into the temp dir to later read all-together.
        file_path = mcfonts.utils.resources.expand_path(file_path)
        shutil.copy(file_path, temp_dir)
        # Make a temporary fake provider file
        with tempfile.NamedTemporaryFile("w+", dir=temp_dir) as temp_file:
            # Overwrite default file, use expanded argument with modifications
            try:
                template_provider["file"] = mcfonts.utils.resources.expand_resource_location(file_path)
            except KeyError as exception:
                if strict:
                    raise exception
            json.dump(template_provider, temp_file, ensure_ascii=False)
        return from_java_font_file(pathlib.Path(temp_file.name), read_colors, strict)


def from_java_ambiguous(path: pathlib.Path, read_colors: bool = False, strict: bool = True) -> mcfonts.MinecraftFont:
    """
    For file paths where the file pointed to is of an unknown type; it could be a JSON, ZIP, or directory.

    This function automatically figures out which function to use, and returns a MinecraftFont.

    :param path:
        The real path to either a file or a folder.
    :param read_colors:
        If True, glyph will be loaded in 'RGBA'.
        If False, loaded in 'LA'.

        RGBA images incur **heavy** time cost.
        Be careful.
    :param strict:
        If True:

        * Bad providers will raise an error.
        * Missing files will raise an error.

        If False:

        * Bad providers will be ignored.
        * Missing files will be skipped.

    :returns: A :class:`mcfonts.MinecraftFont` instance.
    """
    if str(path).endswith(".json"):
        return from_java_font_file(path, read_colors, strict)
    if os.path.isdir(path):
        return from_java_pack_folder(path, read_colors=read_colors, strict=strict)
    # Not a JSON, not a directory, must be a ZIP.
    return from_java_pack_zip(path, read_colors=read_colors, strict=strict)



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Functions for exporting into OpenType XML-representable data (TTX).

Contains functions for transforming and handling OpenType font XML data..
"""
from __future__ import annotations

import datetime
import re
import typing

import lxml.etree

import mcfonts.exceptions
import mcfonts.glyphs
import mcfonts.utils.unicode

XML_FONT_TEMPLATE: bytes = b"""<?xml version="1.0" encoding="UTF-8"?>
<ttFont sfntVersion="OTTO" ttLibVersion="4.34">
    <GlyphOrder>
        <GlyphID id="0" name=".notdef"/>
    </GlyphOrder>
    <head>
        <tableVersion value="1.0"/>
        <fontRevision value="1.0"/>
        <checkSumAdjustment value="0x910274da"/>
        <magicNumber value="0x5f0f3cf5"/>
        <flags value="00000000 00001011"/>
        <unitsPerEm value="1000"/>
        <created value="Mon Jan  1 00:00:00 0000"/>
        <modified value="Mon Jan  1 00:00:00 0000"/>
        <xMin value="-12500"/>
        <yMin value="-12500"/>
        <xMax value="12500"/>
        <yMax value="12500"/>
        <macStyle value="00000000 00000000"/>
        <lowestRecPPEM value="8"/>
        <fontDirectionHint value="2"/>
        <indexToLocFormat value="0"/>
        <glyphDataFormat value="0"/>
    </head>
    <hhea>
        <tableVersion value="0x00010000"/>
        <ascent value="1250"/>
        <descent value="-250"/>
        <lineGap value="250"/>
        <advanceWidthMax value="12500"/>
        <minLeftSideBearing value="0"/>
        <minRightSideBearing value="0"/>
        <xMaxExtent value="1125"/>
        <caretSlopeRise value="1"/>
        <caretSlopeRun value="0"/>
        <caretOffset value="0"/>
        <reserved0 value="0"/>
        <reserved1 value="0"/>
        <reserved2 value="0"/>
        <reserved3 value="0"/>
        <metricDataFormat value="0"/>
        <numberOfHMetrics value="1"/>
    </hhea>
    <maxp>
        <tableVersion value="0x5000"/>
        <numGlyphs value="0"/>
    </maxp>
    <OS_2>
        <!-- The fields 'usFirstCharIndex' and 'usLastCharIndex' will be recalculated by the compiler -->
        <version value="5"/>
        <xAvgCharWidth value="660"/>
        <usWeightClass value="400"/>
        <usWidthClass value="5"/>
        <fsType value="00000000 00000000"/>
        <ySubscriptXSize value="650"/>
        <ySubscriptYSize value="700"/>
        <ySubscriptXOffset value="0"/>
        <ySubscriptYOffset value="140"/>
        <ySuperscriptXSize value="650"/>
        <ySuperscriptYSize value="700"/>
        <ySuperscriptXOffset value="0"/>
        <ySuperscriptYOffset value="480"/>
        <yStrikeoutSize value="50"/>
        <yStrikeoutPosition value="256"/>
        <sFamilyClass value="0"/>
        <panose>
            <bFamilyType value="2"/>
            <bSerifStyle value="0"/>
            <bWeight value="5"/>
            <bProportion value="0"/>
            <bContrast value="0"/>
            <bStrokeVariation value="0"/>
            <bArmStyle value="0"/>
            <bLetterForm value="0"/>
            <bMidline value="6"/>
            <bXHeight value="6"/>
        </panose>
        <ulUnicodeRange1 value="00000000 00000000 00000000 00000000"/>
        <ulUnicodeRange2 value="00000000 00000000 00000000 00000000"/>
        <ulUnicodeRange3 value="00000000 00000000 00000000 00000000"/>
        <ulUnicodeRange4 value="00000000 00000000 00000000 00000000"/>
        <achVendID value="pMCF"/>
        <fsSelection value="00000000 00000000"/>
        <usFirstCharIndex value="32"/>
        <usLastCharIndex value="0"/>
        <sTypoAscender value="875"/>
        <sTypoDescender value="-125"/>
        <sTypoLineGap value="4"/>
        <usWinAscent value="1250"/>
        <usWinDescent value="250"/>
        <!-- This will be filled in automatically -->
        <ulCodePageRange1 value="00000000 00000000 00000000 00000000"/>
        <ulCodePageRange2 value="00000000 00000000 00000000 00000000"/>
        <sxHeight value="600"/>
        <sCapHeight value="1000"/>
        <usDefaultChar value="32"/>
        <usBreakChar value="32"/>
        <usMaxContext value="0"/>
        <usLowerOpticalPointSize value="0"/>
        <usUpperOpticalPointSize value="1"/>
    </OS_2>
    <name></name>
    <cmap>
        <tableVersion version="0"/>
    </cmap>
    <post>
        <formatType value="3.0"/>
        <italicAngle value="0.0"/>
        <underlinePosition value="0"/>
        <underlineThickness value="0"/>
        <isFixedPitch value="0"/>
        <minMemType42 value="0"/>
        <maxMemType42 value="0"/>
        <minMemType1 value="0"/>
        <maxMemType1 value="0"/>
    </post>
    <CFF>
        <major value="1"/>
        <minor value="0"/>
        <CFFFont name="Default">
            <version value="001.000"/>
            <Notice value=""/>
            <FullName value="Default"/>
            <FamilyName value="Default"/>
            <Weight value="Regular"/>
            <isFixedPitch value="0"/>
            <ItalicAngle value="0"/>
            <UnderlinePosition value="0"/>
            <UnderlineThickness value="0"/>
            <PaintType value="0"/>
            <CharstringType value="2"/>
            <FontMatrix value="0.001 0 0 0.001 0 0"/>
            <FontBBox value="0 -250 1125 1250"/>
            <StrokeWidth value="0"/>
            <!-- charset is dumped separately as the 'GlyphOrder' element -->
            <Encoding name="StandardEncoding"/>
            <Private>
                <BlueValues value="0 0 500 500 700 700"/>
                <BlueScale value="0.039625"/>
                <BlueShift value="0"/>
                <BlueFuzz value="1"/>
                <ForceBold value="0"/>
                <StdHW value="125" />
                <StdVW value="125" />
                <StemSnapH value="125 250" />
                <StemSnapV value="125 250" />
                <LanguageGroup value="0"/>
                <ExpansionFactor value="0.06"/>
                <initialRandomSeed value="0"/>
                <defaultWidthX value="750"/>
                <nominalWidthX value="900"/>
            </Private>
            <CharStrings>
                <CharString name=".notdef">
                500 hmoveto 750 -375 -750 vlineto -125 -125 rmoveto 1000 625 -1000 vlineto endchar
                </CharString>
            </CharStrings>
        </CFFFont>
        <GlobalSubrs/>
    </CFF>
    <GDEF>
        <Version value="0x00010000"/>
        <GlyphClassDef/>
    </GDEF>
    <hmtx>
        <mtx name=".notdef" width="750" lsb="0"/>
    </hmtx>
    <DSIG>
        <!-- note that the Digital Signature will be invalid after recompilation! -->
        <tableHeader flag="0x0" numSigs="0" version="1"/>
    </DSIG>
</ttFont>"""
"""
Bytes of a blank TTF XML, with a ``.notdef`` glyph already embedded.
You must call :func:`lxml.etree.XML` with this.
"""

ULUNICODERANGE_BIT_MAP: dict[tuple[int, int], int] = {
    (0x0000, 0x007F): 0,
    (0x0080, 0x00FF): 1,
    (0x0100, 0x017F): 2,
    (0x0180, 0x024F): 3,
    (0x0250, 0x02AF): 4,
    (0x1D00, 0x1D7F): 4,
    (0x1D80, 0x1DBF): 4,
    (0x02B0, 0x02FF): 5,
    (0xA700, 0xA71F): 5,
    (0x0300, 0x036F): 6,
    (0x1DC0, 0x1DFF): 6,
    (0x0370, 0x03FF): 7,
    (0x2C80, 0x2CFF): 8,
    (0x0400, 0x04FF): 9,
    (0x0500, 0x052F): 9,
    (0x2DE0, 0x2DFF): 9,
    (0xA640, 0xA69F): 9,
    (0x0530, 0x058F): 10,
    (0x0590, 0x05FF): 11,
    (0xA500, 0xA63F): 12,
    (0x0600, 0x06FF): 13,
    (0x0750, 0x077F): 13,
    (0x07C0, 0x07FF): 14,
    (0x0900, 0x097F): 15,
    (0x0980, 0x09FF): 16,
    (0x0A00, 0x0A7F): 17,
    (0x0A80, 0x0AFF): 18,
    (0x0B00, 0x0B7F): 19,
    (0x0B80, 0x0BFF): 20,
    (0x0C00, 0x0C7F): 21,
    (0x0C80, 0x0CFF): 22,
    (0x0D00, 0x0D7F): 23,
    (0x0E00, 0x0E7F): 24,
    (0x0E80, 0x0EFF): 25,
    (0x10A0, 0x10FF): 26,
    (0x2D00, 0x2D2F): 26,
    (0x1B00, 0x1B7F): 27,
    (0x1100, 0x11FF): 28,
    (0x1E00, 0x1EFF): 29,
    (0x2C60, 0x2C7F): 29,
    (0xA720, 0xA7FF): 29,
    (0x1F00, 0x1FFF): 30,
    (0x2000, 0x206F): 31,
    (0x2E00, 0x2E7F): 31,
    (0x2070, 0x209F): 32,
    (0x20A0, 0x20CF): 33,
    (0x20D0, 0x20FF): 34,
    (0x2100, 0x214F): 35,
    (0x2150, 0x218F): 36,
    (0x2190, 0x21FF): 37,
    (0x27F0, 0x27FF): 37,
    (0x2900, 0x297F): 37,
    (0x2B00, 0x2BFF): 37,
    (0x2200, 0x22FF): 38,
    (0x2A00, 0x2AFF): 38,
    (0x27C0, 0x27EF): 38,
    (0x2980, 0x29FF): 38,
    (0x2300, 0x23FF): 39,
    (0x2400, 0x243F): 40,
    (0x2440, 0x245F): 41,
    (0x2460, 0x24FF): 42,
    (0x2500, 0x257F): 43,
    (0x2580, 0x259F): 44,
    (0x25A0, 0x25FF): 45,
    (0x2600, 0x26FF): 46,
    (0x2700, 0x27BF): 47,
    (0x3000, 0x303F): 48,
    (0x3040, 0x309F): 49,
    (0x30A0, 0x30FF): 50,
    (0x31F0, 0x31FF): 50,
    (0x3100, 0x312F): 51,
    (0x31A0, 0x31BF): 51,
    (0x3130, 0x318F): 52,
    (0xA840, 0xA87F): 53,
    (0x3200, 0x32FF): 54,
    (0x3300, 0x33FF): 55,
    (0xAC00, 0xD7AF): 56,
    (0x10000, 0x10FFFF): 57,
    (0x10900, 0x1091F): 58,
    (0x4E00, 0x9FFF): 59,
    (0x2E80, 0x2EFF): 59,
    (0x2F00, 0x2FDF): 59,
    (0x2FF0, 0x2FFF): 59,
    (0x3400, 0x4DBF): 59,
    (0x20000, 0x2A6DF): 59,
    (0x3190, 0x319F): 59,
    (0xE000, 0xF8FF): 60,
    (0x31C0, 0x31EF): 61,
    (0xF900, 0xFAFF): 61,
    (0x2F800, 0x2FA1F): 61,
    (0xFB00, 0xFB4F): 62,
    (0xFB50, 0xFDFF): 63,
    (0xFE20, 0xFE2F): 64,
    (0xFE10, 0xFE1F): 65,
    (0xFE30, 0xFE4F): 65,
    (0xFE50, 0xFE6F): 66,
    (0xFE70, 0xFEFF): 67,
    (0xFF00, 0xFFEF): 68,
    (0xFFF0, 0xFFFF): 69,
    (0x0F00, 0x0FFF): 70,
    (0x0700, 0x074F): 71,
    (0x0780, 0x07BF): 72,
    (0x0D80, 0x0DFF): 73,
    (0x1000, 0x109F): 74,
    (0x1200, 0x137F): 75,
    (0x1380, 0x139F): 75,
    (0x2D80, 0x2DDF): 75,
    (0x13A0, 0x13FF): 76,
    (0x1400, 0x167F): 77,
    (0x1680, 0x169F): 78,
    (0x16A0, 0x16FF): 79,
    (0x1780, 0x17FF): 80,
    (0x19E0, 0x19FF): 80,
    (0x1800, 0x18AF): 81,
    (0x2800, 0x28FF): 82,
    (0xA000, 0xA48F): 83,
    (0xA490, 0xA4CF): 83,
    (0x1700, 0x171F): 84,
    (0x1720, 0x173F): 84,
    (0x1740, 0x175F): 84,
    (0x1760, 0x177F): 84,
    (0x10300, 0x1032F): 85,
    (0x10330, 0x1034F): 86,
    (0x10400, 0x1044F): 87,
    (0x1D000, 0x1D0FF): 88,
    (0x1D100, 0x1D1FF): 88,
    (0x1D200, 0x1D24F): 88,
    (0x1D400, 0x1D7FF): 89,
    (0xF0000, 0xFFFFD): 90,
    (0x100000, 0x10FFFD): 90,
    (0xFE00, 0xFE0F): 91,
    (0xE0100, 0xE01EF): 91,
    (0xE0000, 0xE007F): 92,
    (0x1900, 0x194F): 93,
    (0x1950, 0x197F): 94,
    (0x1980, 0x19DF): 95,
    (0x1A00, 0x1A1F): 96,
    (0x2C00, 0x2C5F): 97,
    (0x2D30, 0x2D7F): 98,
    (0x4DC0, 0x4DFF): 99,
    (0xA800, 0xA82F): 100,
    (0x10000, 0x1007F): 101,
    (0x10080, 0x100FF): 101,
    (0x10100, 0x1013F): 101,
    (0x10140, 0x1018F): 102,
    (0x10380, 0x1039F): 103,
    (0x103A0, 0x103DF): 104,
    (0x10450, 0x1047F): 105,
    (0x10480, 0x104AF): 106,
    (0x10800, 0x1083F): 107,
    (0x10A00, 0x10A5F): 108,
    (0x1D300, 0x1D35F): 109,
    (0x12000, 0x123FF): 110,
    (0x12400, 0x1247F): 110,
    (0x1D360, 0x1D37F): 111,
    (0x1B80, 0x1BBF): 112,
    (0x1C00, 0x1C4F): 113,
    (0x1C50, 0x1C7F): 114,
    (0xA880, 0xA8DF): 115,
    (0xA900, 0xA92F): 116,
    (0xA930, 0xA95F): 117,
    (0xAA00, 0xAA5F): 118,
    (0x10190, 0x101CF): 119,
    (0x101D0, 0x101FF): 120,
    (0x102A0, 0x102DF): 121,
    (0x10280, 0x1029F): 121,
    (0x10920, 0x1093F): 121,
    (0x1F030, 0x1F09F): 122,
    (0x1F000, 0x1F02F): 122,
}
"""
A mapping of (block codepoint start, block codepoint end) to ulUnicodeRange bit.
It is best to store a single 128 bit number and then split it up for processing later.
See https://learn.microsoft.com/en-us/typography/opentype/spec/os2#ur.
"""

EXCLUDED_FONT_NAME_CHARS = re.compile(r"[{}\[\]() <>%/]")


@typing.no_type_check
def allocate_character(font_xml: lxml.etree._Element, character: str) -> None:
    """
    Allocate a Unicode character in a font.

    This doesn't assign any character data or widths.

    :param font_xml: The font XML.
    :param character: A single character.
    :raises mcfonts.exceptions.GlyphLimitError: If there are more than 65,535 glyphs in the font; can't add more.
    """

    if (metrics_num := int((metrics := font_xml.find("hhea/numberOfHMetrics")).get("value"))) >= 65535:
        raise mcfonts.exceptions.GlyphLimitError()
    metrics.set("value", str(metrics_num + 1))

    num_glyphs = font_xml.find("maxp/numGlyphs")
    num_glyphs.set("value", str(int(num_glyphs.get("value")) + 1))

    cmap = font_xml.find("cmap")
    codepoint = ord(character)
    uni = f"u{codepoint:04X}"
    make_cmap(
        font_xml,
        (codepoint <= 0xFFFF and (cmap.find("cmap_format_4")) is None),  # BMP
        (codepoint < 0xFFFF and (cmap.find("cmap_format_12") is None)),  # above BMP
    )
    lxml.etree.SubElement(
        cmap.find("cmap_format_12"), "map", {"code": f"0x{codepoint:X}", "name": uni, "language": "0"}
    )

    if codepoint <= 0xFFFF:
        lxml.etree.SubElement(
            cmap.find("cmap_format_4"), "map", {"code": f"0x{codepoint:X}", "name": uni, "language": "0"}
        )

    lxml.etree.SubElement(font_xml.find("GDEF/GlyphClassDef"), "ClassDef", {"glyph": uni, "class": "1"})

    lxml.etree.SubElement(font_xml.find("GlyphOrder"), "GlyphID", {"name": uni})


@typing.no_type_check
def make_cmap(font_xml: lxml.etree._Element, format_4: bool = True, format_12: bool = True) -> None:
    """
    Insert cmap subtables in `font_xml`. These are used for mapping glyph IDs to Unicode codepoints.

    If `format_4`, cmap_format_4 will be added. If `format_12`, cmap_format_12 will be added.

    * Format 4: https://learn.microsoft.com/en-us/typography/opentype/spec/cmap#format-4-segment-mapping-to-delta-values
    * Format 12: https://learn.microsoft.com/en-us/typography/opentype/spec/cmap#format-12-segmented-coverage

    :param font_xml: A font XML.
    :param format_4:
        Whether to insert a cmap format 4 subtable.
        See https://learn.microsoft.com/en-us/typography/opentype/spec/cmap#format-4-segment-mapping-to-delta-values.
    :param format_12:
        Whether to insert a cmap format 12 subtable.
        See https://learn.microsoft.com/en-us/typography/opentype/spec/cmap#format-12-segmented-coverage.
    """
    cmap = font_xml.find("cmap")
    if format_12 and cmap.find("cmap_format_12") is None:
        lxml.etree.SubElement(
            cmap,
            "cmap_format_12",
            {
                "platformID": "0",
                "platEncID": "4",
                "language": "0",
                "length": "4120",
                "nGroups": "323",
                "format": "12",
                "reserved": "0",
            },
        )
    if format_4 and cmap.find("cmap_format_4") is None:
        lxml.etree.SubElement(
            cmap,
            "cmap_format_4",
            {
                "platformID": "3",
                "platEncID": "4",
                "language": "0",
            },
        )


@typing.no_type_check
def set_program_to_character(
    font_xml: lxml.etree._Element, program: list[str | int], character: str, replace: bool = True
):
    """
    Set a program to a character.

    This is how character data is added to the font.
    The character must be allocated already; that isn't done in this function.

    If the character isn't in the font, add it.
    If it's and ``replace`` is True, set it to the new character data.
    Otherwise, do nothing.

    :param font_xml: The font XML.
    :param program: A list of strings of the glyph's program.
    :param character: A single character.
    :param replace: If the character already has data, overwrite.
    :returns: Nothing, font is modified in-place.
    :raises GlyphLimitError: If there are more than 65,535 glyphs in the font; can't add more.
    """
    uni = f"u{ord(character):04X}"
    charstrings = font_xml.find("CFF/CFFFont/CharStrings")
    charstring_xml = font_xml.find(f"{charstrings}/CharString[@name='{uni}']")
    if charstring_xml is None:
        # Doesn't exist
        lxml.etree.SubElement(charstrings, "CharString", {"name": uni}).text = " ".join(str(x) for x in program)
        lxml.etree.SubElement(font_xml.find("hmtx"), "mtx", {"name": uni, "width": str(program[0]), "lsb": "0"})
    elif replace:
        # Exists already in the charstrings, replace it too
        charstring_xml.text = " ".join(str(x) for x in program)
        font_xml.find(f"hmtx/mtx[@name='{uni}']").set("width", str(program[0]))


@typing.no_type_check
def set_space_to_character(font_xml: lxml.etree._Element, character: str, width: int, replace: bool = True) -> None:
    """
    Add/set a space character to the font, only defining its width.

    The character must be allocated already.
    That isn't done in this function.

    If the glyph isn't in the font, add it.
    If it's and ``replace`` is True, set it to the new value.
    Otherwise, do nothing.

    :param font_xml: The font XML.
    :param character: A single character.
    :param width: The width of the space, unscaled.
    :param replace: If the character already has data, overwrite.
    :returns: Nothing, font is modified in-place.
    :raises GlyphLimitError: If there are more than 65,535 glyphs in the font; can't add more.
    """
    uni = f"u{ord(character):04X}"
    charstrings = font_xml.find("CFF/CFFFont/CharStrings")
    width = abs(width) * 125

    if (charstring_xml := font_xml.find(f"{charstrings}/CharString[@name='{uni}']")) is None:
        # Doesn't exist
        lxml.etree.SubElement(charstrings, "CharString", {"name": uni}).text = f"{width} endchar"
        lxml.etree.SubElement(font_xml.find("hmtx"), "mtx", {"name": uni, "width": str(width), "lsb": "0"})
    elif replace:
        # Exists already in the charstrings, replace it too
        charstring_xml.text = f"{width} endchar"
        font_xml.find(f"hmtx/mtx[@name='{uni}']").set("width", str(width))


@typing.no_type_check
def add_namerecord_to_font(font_xml: lxml.etree._Element, data: str, name_id: int) -> None:
    """
    Set a namerecord in a font.

    This doesn't check if such a namerecord already exists; this will add new namerecords.

    * 0 -> Copyright
    * 1 -> Font family
    * 2 -> Font subfamily
    * 3 -> Unique font ID
    * 4 -> Full font name, ID 1 + 2
    * 5 -> Version: "Version maj.min"
    * 6 -> PostScript name
    * 7 -> Trademark
    * 8 -> Manufacturer
    * 9 -> Designer
    * 10 -> Descriptions
    * ... see more at https://docs.microsoft.com/en-us/typography/opentype/spec/name#name-ids.

    All are encoded at (0, 4) and (3, 1), Unicode 2.0+ full.

    :param font_xml: The font XML.
    :param data: Whatever string to add.
    :param name_id: The ID of the namerecord.
    :returns: Nothing, font is modified in-place.
    """
    unicode_platenc = "3"  # Default value for Unicode BMP only.
    windows_platenc = "1"  # Default value for Windows BMP only.

    if any(ord(character) > 0xFFFF for character in data):  # Full Unicode
        unicode_platenc = "4"
        windows_platenc = "10"
    name_elements = font_xml.find("name")
    name_id = str(name_id)
    lxml.etree.SubElement(
        name_elements,
        "namerecord",
        {"platformID": "0", "platEncID": unicode_platenc, "nameID": name_id, "langID": "0x0"},
    ).text = data  # Unicode
    lxml.etree.SubElement(
        name_elements,
        "namerecord",
        {"platformID": "1", "platEncID": "32", "nameID": name_id, "langID": "0x0"},
    ).text = data.encode(
        "latin-1", errors="backslashreplace"
    )  # Macintosh (latin-1 only)
    lxml.etree.SubElement(
        name_elements,
        "namerecord",
        {"platformID": "3", "platEncID": windows_platenc, "nameID": name_id, "langID": "0x409"},
    ).text = data  # Windows


@typing.no_type_check
def set_cfffont_name(font_xml: lxml.etree._Element, font_name: str, family_name: str) -> None:
    """
    Set the font's appropriate CFF name in 'CFF ' tables.

    This is different from changing the 'name' tables, see :func:`add_namerecord_to_font` instead.

    :param font_xml: The font XML.
    :param font_name: The name of the font (doesn't include "Regular" or related weights).
    :param family_name: The family name of the font (doesn't include "Regular" or related weights).
    :returns: Nothing, font is modified in-place.
    """
    cfffont = font_xml.find("CFF/CFFFont")
    cfffont.set("name", sanitize_font_name(font_name))
    cfffont.find("FullName").set("value", font_name)
    cfffont.find("FamilyName").set("value", family_name)
    mcfonts.logger.debug(f"exporting: set CFF font name to {font_name}, family {family_name}")


@typing.no_type_check
def set_font_times(font_xml: lxml.etree._Element) -> None:
    """
    Set the font's created and modified times, in the format of ``%a %b %d %X %Y``.

    :param font_xml: The font XML.
    :returns: Nothing, font is modified in-place.
    """
    time = datetime.datetime.now().strftime("%a %b %d %X %Y")
    font_xml.find("head/created").set("value", time)
    font_xml.find("head/modified").set("value", time)


@typing.no_type_check
def set_notdef_in_font(font_xml: lxml.etree._Element, program: list[str]) -> None:
    """
    Set the ``.notdef`` character of the font.

    .. note::
        The default font template already has a default notdef, so use this for setting it to something else.

    :param font_xml: The font XML.
    :param program: A list of strings of the glyph's program.
    :returns: Nothing, font is modified in-place.
    """
    font_xml.find("CFF/CFFFont/CharStrings/CharString[@name='.notdef']").text = " ".join(str(x) for x in program)
    font_xml.find("hmtx/mtx[@name='.notdef']").set("width", str(program[0]))
    mcfonts.logger.debug("exporting: set notdef")


@typing.no_type_check
def set_font_name(font_xml: lxml.etree._Element, font_name: str) -> None:
    """
    Set the font's name to `font_xml` in all appropriate places.

    This is different from :func:`set_cfffont_name`, which sets the name for only CFF tables.

    :param font_xml: The font XML.
    :param font_name: The name of the font.
    """
    font_name = font_name.encode("ascii", "ignore").decode("utf-8")  # Ignore anything not in ASCII
    add_namerecord_to_font(font_xml, font_name, 1)
    set_cfffont_name(font_xml, font_name, font_name)
    add_namerecord_to_font(font_xml, "Regular", 2)
    add_namerecord_to_font(font_xml, f"{font_name} Regular: {hash(font_xml)}", 3)
    add_namerecord_to_font(font_xml, f"{font_name} Regular", 4)
    add_namerecord_to_font(
        font_xml,
        f"Version 1.0; {datetime.date.strftime(datetime.date.today(), '%B %d, %Y')}",
        5,
    )
    add_namerecord_to_font(font_xml, sanitize_font_name(font_name), 6)


def sanitize_font_name(font_name: str) -> str:
    """
    Ensure ``sanitized_font_name`` is a valid PostScript font name.

    A PostScript font name can't:

    * Contain ``(){}[]<;>%/`` or space
    * Be longer than 63 characters
    * Have non-ASCII characters

    >>> sanitize_font_name("\u2600 This is a really long name!")
    '_This_is_a_really_long_name!'
    >>> sanitize_font_name("(secret info) [do not leak]")
    '_secret_info___do_not_leak_'
    >>> sanitize_font_name("a really long name with over 63 characters I guarantee it you won't believe it")
    'a_really_long_name_with_over_63_characters_I_guarantee_it_you_w'

    :param font_name: The font name to sanitize.
    :returns: A valid PostScript font name.
    """
    return EXCLUDED_FONT_NAME_CHARS.sub("_", font_name)[:63].encode("ascii", "ignore").decode("ascii")


def generate_ulunicoderange(characters: set[str]) -> list[str]:
    """
    Given a set of characters, return a list of ulUnicodeRanges, with bits set appropriately.

    See https://learn.microsoft.com/en-us/typography/opentype/spec/os2#ur.

    >>> generate_ulunicoderange({"A", "B", "\u2600", "\u1BFF", "\uFFFD"})
    [
        '00000000 00000000 00000000 00000000',
        '00000000 00000000 00000000 00100000',
        '00000000 00000000 01000000 00000000',
        '00000000 00000000 00000000 00000001',
    ]

    :param characters: A set of characters.
    :return: A list of ulUnicodeRanges, as strings, in order from 1 to 4.
    """
    ul = 0
    for character in characters:
        codepoint = ord(character)
        for unicode_range, bit in ULUNICODERANGE_BIT_MAP.items():
            if unicode_range[0] <= codepoint <= unicode_range[1]:
                # Set the `bit`th bit in integer `ul`.
                # See https://stackoverflow.com/a/28360656.
                ul ^= (-1 ^ ul) & (1 << bit)
    ul_binary_string = f"{ul:0128b}"
    ul_finished = []
    # Split string into lengths, see https://stackoverflow.com/a/13673133.
    for elem in (ul_binary_string[i : i + 32] for i in range(0, 128, 32)):
        ul_finished.append(" ".join([elem[i : i + 8] for i in range(0, 32, 8)]))
    return ul_finished[::-1]


@typing.no_type_check
def set_ulunicoderange(font_xml: lxml.etree._Element, characters: set[str]) -> None:
    """
    Generate the ulUnicodeRanges corresponding to `characters` and set them in `font_xml`.

    :param font_xml: A font XML.
    :param characters: A set of characters to generate ulUnicodeRanges for.
    """
    ul = generate_ulunicoderange(characters)
    os2 = font_xml.find("OS_2")
    os2.find("ulUnicodeRange1").set("value", ul[0])
    os2.find("ulUnicodeRange2").set("value", ul[1])
    os2.find("ulUnicodeRange3").set("value", ul[2])
    os2.find("ulUnicodeRange4").set("value", ul[3])


def include_glyph(
    font_xml: lxml.etree._Element, glyph: mcfonts.glyphs.Glyph, character: str, allocated_characters: set[str]
) -> None:
    """
    Include `glyph` into the font XML.

    Different Glyph subtypes have different methods of being imported;
    this function centralizes those varying methods.

    :param font_xml: A font XML.
    :param glyph: An instance of :class:`mcfont.glyphs.Glyph`.
    :param character: The character to assign this glyph to.
    :param allocated_characters: A set of the characters already included in the font XML.
    """
    if character not in allocated_characters:
        allocate_character(font_xml, character)
    if isinstance(glyph, mcfonts.glyphs.SpaceGlyph):
        set_space_to_character(font_xml, character, min(glyph.get_width(), 512))
    else:
        if program := glyph.get_program():  # If there's a program.
            set_program_to_character(font_xml, program, character, False)
        else:
            # Nothing here; it was a space.
            set_space_to_character(font_xml, character, 1, False)



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""JSON schemas for providers. Dynamically loaded from ``mcfonts/schemas/``."""
from __future__ import annotations

import json
import pathlib

with pathlib.Path(__file__).parent.parent.joinpath("schemas/provider.bitmap.schema.json").open() as __x:
    SCHEMA_PROVIDER_BITMAP = json.load(__x)
    """A schema to validate "bitmap" providers against."""

with pathlib.Path(__file__).parent.parent.joinpath("schemas/provider.space.schema.json").open() as __x:
    SCHEMA_PROVIDER_SPACE = json.load(__x)
    """A schema to validate "space" providers against."""

with pathlib.Path(__file__).parent.parent.joinpath("schemas/provider.legacy_unicode.schema.json").open() as __x:
    SCHEMA_PROVIDER_LEGACY_UNICODE = json.load(__x)
    """A schema to validate "legacy_unicode" providers against."""

with pathlib.Path(__file__).parent.parent.joinpath("schemas/provider.ttf.schema.json").open() as __x:
    SCHEMA_PROVIDER_TTF = json.load(__x)
    """A schema to validate "ttf" providers against."""

with pathlib.Path(__file__).parent.parent.joinpath("schemas/provider.unihex.schema.json").open() as __x:
    SCHEMA_PROVIDER_UNIHEX = json.load(__x)
    """A schema to validate "unihex" providers against."""

with pathlib.Path(__file__).parent.parent.joinpath("schemas/provider.reference.schema.json").open() as __x:
    SCHEMA_PROVIDER_REFERENCE = json.load(__x)
    """A schema to validate "reference" providers against."""

with pathlib.Path(__file__).parent.parent.joinpath("schemas/providers.schema.json").open() as __x:
    SCHEMA_PROVIDERS = json.load(__x)
    """A schema to validate the base font providers dictionary against."""



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A space provider, defining simple widths for characters."""
import collections.abc
import os
import typing

import unirange

import mcfonts.constants
import mcfonts.glyphs
import mcfonts.providers
import mcfonts.utils


class SpaceProvider(mcfonts.providers.base.Provider):
    """
    The ``space`` provider defines only the width of a character.

    It's the simplest provider: it only contains the ``"advances"`` map.

    The keys of the advances are the characters as 1-length strings, and the values are the widths as integers.
    In mcfonts only, uniranges can also be used as the keys.

    The default width for ``U+0020 SPACE`` is 4.
    """

    def __init__(self, provider: mcfonts.utils.types.TypedProviderSpace, origin: os.PathLike[str] | str) -> None:
        self.contents = provider
        self.glyphs: dict[str, mcfonts.glyphs.SpaceGlyph] = dict(self.construct_glyphs())
        for key, value in provider.get("advances", {}).items():
            for new_character in unirange.unirange_to_characters(key):
                # Expand the uniranges and copy over fields.
                self.contents["advances"][new_character] = value
        super().__init__("space", origin, set(provider["advances"].keys()))
        self.validate()

    def pretty_print(self) -> str:
        return f"{self.get_icon()} space: {len(self.get_contents().get('advances', [])):,} widths"

    def validate(self) -> None:
        return None  # No specific validation checks yet.

    def print_info(self) -> None:
        super().print_info()

        if len(self.contents["advances"]) < 1:
            # There's no "advances" dictionary.
            print("No advances.")
        else:
            advances = self.contents["advances"]
            print(f"Advances: ({len(advances):,})")
            for space_character, width in advances.items():
                print(f"\tCharacter {mcfonts.utils.unicode.pretty_print_character(space_character)}: {width}")

    def construct_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.SpaceGlyph]]:
        for character, width in self.contents["advances"].items():
            yield character, mcfonts.glyphs.SpaceGlyph(int(width))

    def yield_characters(self) -> collections.abc.Iterable[str]:
        yield from self.contents["advances"]

    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.SpaceGlyph]]:
        yield from self.glyphs.items()

    def get_icon(self) -> str:
        return "␣"

    def get_pack_versions(self) -> tuple[int, None]:
        return 9, None

    def yield_glyphs_in_unirange(
        self, unirange_notation: str
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.SpaceGlyph | None]]:
        for character in unirange.unirange_to_characters(unirange_notation):
            yield character, self.glyphs.get(character)

    def get_contents(self) -> mcfonts.utils.types.TypedProviderSpace:
        return self.contents


def to_glyph_sizes(advances: dict[str, int], default_width: tuple[int, int] = (0, 14)) -> bytearray:
    r"""
    Create a glyph_sizes.bin bytearray from a template of characters and their starting and ending positions.

    >>> to_glyph_sizes({"\0": 15}, (0, 0))
    bytearray(b'\x0e\x00\x00\x00...')

    :param advances: A dictionary of ``{character: width}``.
    :param default_width:
        The width to fall back to if `advances` doesn't define one for a character.
    :returns: Bytearray of glyph_sizes.bin.
    """
    glyphsizes = bytearray((default_width[0] * 16 + default_width[1]).to_bytes(1, "big") * 65536)
    for character, width in advances.items():
        if (codepoint := ord(character)) > 0xFFFF:
            # Can't pack characters higher than the BMP.
            mcfonts.logger.warning(f"Cannot include character {character} in glyph_sizes; codepoint is above U+FFFF.")
            continue
        # Ensure the high and low bits are in the correct 0-F range.
        if 0 > width < 15:
            raise ValueError("Width must be within 0 to 15")
        glyphsizes[codepoint] = width
    return glyphsizes
